Drupal.behaviors.entity_mesh = {
  attach: function (context, settings) {
    once('entityMesh', '#entity-mesh-graph', context).forEach(
      function (element) {
        element.classList.add('processed');
        if (settings.entity_mesh.data) {
          graphGenerate(settings.entity_mesh.data, settings.entity_mesh.settings);
        } else {
          graphGenerate(entity_mesh_data, settings.entity_mesh.settings);
        }
      }
    );
  }
};

function graphGenerate(data, settings) {
  // Initialize objects to ensure they are reset for each request
  let selector = '#entity-mesh-graph',
    nodes = [],
    links = [],
    types = [],
    selectedNode = null;
  let filterTypes = [];

  // Assign new data to the objects
  nodes = data.nodes;
  links = data.links;
  types = data.types;

  // Init Graph:
  let graph = document.querySelector(selector);
  let width = window.innerWidth - 100;
  let height = 450;
  if (settings.fullHeight) {
    height = window.innerHeight - 300;
  }

  // Init Color helper:
  let color = d3.scaleOrdinal(types, d3.schemeCategory10);

  // Generate a D3 SVG graph:
  const svg = d3
    .select(selector)
    .append("svg")
    .attr("width", width)
    .attr("height", height);

  // Init simulation:
  // @see https://github.com/d3/d3-force
  let simulation = d3.forceSimulation()
    .force("center", d3.forceCenter(width / 2, height / 2))
    .force("link", d3.forceLink().id(function(d) { return d.id;}))
    .force("charge", d3.forceManyBody().strength(-200).distanceMax(Math.min(width,height)/2))
    .force("collide", d3.forceCollide().strength(-200))
    .force("y", d3.forceY(0))
    .force("x", d3.forceX(0));

  // Zoom wrapper:
  var wrapper = svg.append("g")
    .attr("class", "zoom-wrapper");

  // Arrow marker:
  let marker = wrapper
    .append("defs")
    .append("marker")
    .attr("id", "arrow")
    .attr("viewBox", "0 -5 10 10")
    .attr('fill', '#DDD')
    .attr("refX", 25)
    .attr("refY", 0)
    .attr("markerWidth", 6)
    .attr("markerHeight", 6)
    .attr("orient", "auto");

  marker
    .append("path")
    .attr("d", "M0,-5L10,0L0,5")
    .attr("class", "arrow-head");

  const link = wrapper.append("g")
    .attr("class", "lines")
    .selectAll("line")
    .data(links)
    .enter()
    .append("line")
    .attr("stroke", "#DDD")
    .attr("stroke-width", 2)
    .attr("marker-end", "url(#arrow)");

  // Nodes list:
  let node = wrapper.append("g")
    .attr("class", "nodes")
    .selectAll("g")
    .data(nodes)
    .enter()
    .append("g")
    .attr("class", "node");

  // Circles (on nodes):
  let circles = node
    .append("circle")
    .attr("r", function(d) {
      let baseRadius = 5; // Base radius for all circles.
      let variableSize = getLinkCount(d.id) * (Math.log10(getLinkCount(d.id)) + 1);

      // Handle invalid or NaN values.
      if (isNaN(variableSize)) {
        variableSize = 1;
      }

      // Add the variable size to the base radius.
      let computedRadius = baseRadius + variableSize;

      // Cap the radius to a maximum value (e.g., 50).
      let maxRadius = 50;
      return Math.min(computedRadius, maxRadius);
    })
    .attr("fill", function(d) {
      return color(d.type); // Set the fill color dynamically.
    });

  // Init Drag handlers:
  let drag_handler = d3.drag()
    .on("start", dragstarted)
    .on("drag", dragged)
    .on("end", dragended);

  // Attach nodes to Drag handler:
  drag_handler(node);

  // Add title to nodes:
  node.append("title")
    .text(function(d) {
      return d.label;
    });

  // Add labels for the number of links
  let linkCounts = node
    .append('text')
    .text(function(d) {
      let count = getLinkCount(d.id);
      return count > 1 ? count : '';
    })
    .attr('font-size', '10px')
    .attr('fill', 'white')
    .attr('text-anchor', 'middle')
    .attr('y', 3);

  // Count of links for a node (as target)
  function getLinkCount(nodeId) {
    let count = links.filter(link => link.target === nodeId).length;
    return count > 0 ? count : '';
  }

  node.on("click", toggleConnections);

  function getConnectedNodes(nodeId, depth = 0) {
    const connectedNodes = new Set();
    const visitedNodes = new Set();

    const traverse = (currentNodeId, depth) => {
        visitedNodes.add(currentNodeId);

        links.forEach((link) => {
            // New node target (first level only):
            if (link.source.id === currentNodeId && !visitedNodes.has(link.target.id)) {
                connectedNodes.add(link.target.id);
                if (depth > 0) {
                  traverse(link.target.id, depth - 1);
                }
            }
            // New node that points to the target node (first level only)
            else if (link.target.id === currentNodeId && !visitedNodes.has(link.source.id)) {
                connectedNodes.add(link.source.id);
                if (depth > 0) {
                  traverse(link.source.id, depth - 1);
                }
            }
        });
    };

    traverse(nodeId, depth);

    return connectedNodes;
  }

  // Load nodes on simulation:
  simulation.nodes(nodes)
    .alphaTarget(0)
    .on("tick", ticked);

  // Load links on simulation:
  simulation.force("link")
    .links(links);

  var zoom_handler = d3.zoom()
    .on("zoom", zoom_actions);

  zoom_handler(svg);

  // Legend
  let legend = svg.append("g")
    .attr("class", "legend")
    .attr("transform", "translate(20, 20)");

  let legendItems = legend.selectAll(".legend-item")
    .data(types)
    .enter()
    .append("g")
    .attr("x", 20)
    .attr("y", 10)
    .attr("class", "legend-item")
    .attr("transform", (d, i) => `translate(0, ${i * 30})`);

  legendItems.append("rect")
    .attr("width", 10)
    .attr("height", 10)
    .style("fill", d => color(d));

  legendItems.append("text")
    .attr("x", 20)
    .attr("y", 10)
    .text(d => d)
    .on('click', toggleLegends);

  function toggleLegends(event, clickedLegend) {

    const index = filterTypes.indexOf(clickedLegend);
    if (index === -1) {
      filterTypes.push(clickedLegend);
      d3.select(event.currentTarget).classed('selected', true);

    } else {
      filterTypes.splice(index, 1);
      d3.select(event.currentTarget).classed('selected', false);
    }

    if (filterTypes.length == types.length) {
      filterTypes = []
      // @todo remove
      legendItems.classed('selected', false);
    }

    filterNodes();
  }

  function toggleConnections(event, clickedNode) {
    node.classed('selected', false);

    // Revert selection:
    if (clickedNode === selectedNode) {
      node.attr("display", "inline");
      link.attr("display", "inline");

      selectedNode = null;
      return;
    }

    selectedNode = clickedNode;
    // Apply the 'selected' class to the clicked node
    d3.select(event.currentTarget).classed('selected', true);

    filterNodes();
  }

  function filterNodes() {
    // Hide All:
    node.attr("display", "none");
    link.attr("display", "none");

    let nodesVisible = node;
    let linksVisible = link

    // Filter by Selected node:
    if (selectedNode !== null) {
      let connectedNodes = getConnectedNodes(selectedNode.id, 0);
      // Remove all nodes not connected but current one:
      nodesVisible = node.filter((d) => connectedNodes.has(d.id) || d.id === selectedNode.id);
      // Remove all links from nodes not in the list but current one:
      linksVisible = link.filter((d) => d.source.id == selectedNode.id || d.target.id == selectedNode.id || connectedNodes.has(d.source.id) && connectedNodes.has(d.target.id));
    }

    // Filter by Type:
    if (filterTypes.length > 0) {
      nodesVisible = nodesVisible.filter((d) => filterTypes.indexOf(d.type) !== -1);
      linksVisible = linksVisible.filter((d) => filterTypes.indexOf(d.source.type) !== -1 && filterTypes.indexOf(d.target.type) !== -1);
    }

    // Show Visible ones:
    nodesVisible.attr("display", "inline");
    linksVisible.attr("display", "inline");
  }



  //Zoom functions
  function zoom_actions(event){
    wrapper.attr("transform", event.transform);
  }

  function ticked() {
    node.attr("transform", d => `translate(${d.x},${d.y})`);
    link
      .attr("x1", (d) => d.source.x)
      .attr("y1", (d) => d.source.y)
      .attr("x2", (d) => d.target.x)
      .attr("y2", (d) => d.target.y);
  }

  function dragstarted(event, d) {
    if (!event.active) simulation.alphaTarget(0.3).restart();
    d.fx = d.x;
    d.fy = d.y;
  }

  function dragged(event, d) {
    d.fx = event.x;
    d.fy = event.y;
  }

  function dragended(event, d) {
    if (!event.active) simulation.alphaTarget(0);
    d.fx = null;
    d.fy = null;
  }

}
